/* S.R. Delwiche, USDA/ARS/BA/ANRI/ISL, Building 303, BARC-East, Beltsville, MD 20705-2350 USA, delwiche@ba.ars.usda.gov
  Program filename: PLS_with_pretreatments_and_validation_and_compression23.sas
  Date last modified: March 16, 2004                                                                                   */

/*  New features: Includes validation on a multifile that is similar in structure to the calibration file
                  and allows for within-spectrum compression by averaging.*/

/* ........................................... 1 start .......................................... */

/* ----- Macro variables that seldom need changing ----- */
%Let  progname =PLS_with_pretreatments_and_validation_and_compression23.sas; /* the program's filename, for printout purposes */
%Let      maxs = 1050; /* maximum number of spectral data points */
%Let      maxb = 4200; /* 4*maxs, i.e., assumes 4 bytes per input spectral value */
%Let numtables =   10; /* this value must exactly match the number of tables in SGfile */
%Let    maxfac =   15; /* upper limit for the number of PLS factors */
%Let     bzero =    0; /* initial dummy value of constant term of PLS solution regression vector */

/* ----- Macro variables that specify input files ----- */
%Let      SGfile =d:\pls\sas\savgolgap.prn; /* the 1st 7 Savitzky-Golay & 3 gap convolution tables */
%Let    specfile =d:\wheatstress\data\SASPLS\prosdssweather_rep1.spc;  /* must be Grams multifile (newstyle, even x spacing, constant #pts per spectrum) */
%Let    chemfile =d:\wheatstress\data\SASPLS\sdss_rep1.txt; /* corresponding constituent file, one column, with samples in same order as *.spc file */
%Let valspecfile =d:\wheatstress\data\SASPLS\prosdssweather_rep2_exported.spc; /* same structure as specfile */
%Let valchemfile =d:\wheatstress\data\SASPLS\sdss_rep2.txt; /* same structure as chemfile */

/* ----- Macro variables that specify output files ----- */
%Let   SASstyleoutfile =d:\wheatstress\data\SASPLS\pls_23a.prn; /* for optimal #factors based on minimum PRESS value of cross validation */
%Let Gramsstyleoutfile =d:\wheatstress\data\SASPLS\pls_23b.prn; /* for optimal #factors by F-test of PRESS values of cross validation */

/* ----- Macro variables that truncate and compress the spectral region before any other preprocessing ----- */
%Let           leftcut = 0; /* the number of spectral points on left end to cut */
%Let          rightcut = 0; /* ditto for right end */
%Let compressionfactor = 2; /* must be a positive integer, compresses truncated spectra by this factor by averaging */

/* ----- Macro variables that set loop limits in the Macros - MSCnoSGloop and SGloop ----- */
%Let pathbegin = 0; /* lower loop limit for multiplicative scatter correction, must be 0 (for no msc) or 1 (for msc) */
%Let   pathend = 1; /* upper loop limit, see above */

/* ----- Macro variables that set loop limits in the Macros - SGloop and SNVloop ----- */
%Let  firstSGtable =  4; /* must be 1, 2, 3, 4, 5, 6, 7, 8, 9, or 10 */
%Let secondSGtable =  6; /* see above */
%Let  thirdSGtable = 10; /* see above */
%Let    tablebegin =  1; /* lower loop limit for SG tables, must be 1, 2, or 3 (corresponding to firstSGtable, secondSGtable, and thirdSGtable, respectively) */
%Let      tableend =  3; /* upper loop limit, see above */
%Let   windowbegin =  5; /* lower loop limit for Savitzky-Golay convolution window, must be 5,7,...,25 */
%Let     windowend = 25; /* upper loop limit, see above */

/* ----- Macro variables that set loop limits in the Macros - SNVnoSGloop and SNVloop ----- */
%Let detrendbegin = 0; /* lower loop limit for detrending after SNV, must be 0 (for no detrend) or 1 (for detrend) */
%Let   detrendend = 1; /* upper loop limit, see above

/* ----- Macro variables that set loop limits in the Macros - MSCnoSGloop, SGloop, SNVnoSGloop, and SNVloop ----- */
%Let varscalebegin = 0; /* lower loop limit for variance scaling, must be 0 (for no vs) or 1 (for vs) */
%Let   varscaleend = 0; /* upper loop limit, see above */

/* ........................................... 1 end ............................................ */

/* Macro that reads a single newstyle Grams multifile (evenly spaced Y values) */

%MACRO getspec;

/* ........................................... 2 start .......................................... */

  data specdata;
  array x(&maxb) x1-x&maxb;
  array y(&maxs) y1-y&maxs;
  infile "&specfile" recfm=n;
  input @2 spcversion ibr1.;
  if spcversion < 0 then spcversion = spcversion + 256;
  if spcversion ^= 75 then abort;
  input @5 numpoints1 ibr1. numpoints2 ibr1. @24 +1 numspec1 ibr1. numspec2 ibr1. @512;
  if numpoints1 < 0 then numpoints1 = numpoints1 + 256;
  if numpoints2 < 0 then numpoints2 = numpoints2 + 256;
  numpoints = numpoints2*256 + numpoints1;
  leftedge = &leftcut + 1;
  attrib charleft format = $CHAR.;
  charleft = leftedge;
  charleft = compress(charleft);
  call symput('leftspec',charleft);
  rightedge = numpoints - &rightcut;
  attrib charright format = $CHAR.;
  charright = rightedge;
  charright = compress(charright);
  call symput('rightspec',charright);
  newnumpoints = numpoints - &leftcut - &rightcut;
  attrib charnewnumpoints format = $CHAR.;
  charnewnumpoints = newnumpoints;
  call symput('newpoints',charnewnumpoints);
  if numspec1 < 0 then numspec1 = numspec1 + 256;
  if numspec2 < 0 then numspec2 = numspec2 + 256;
  numrecords = numspec2*256 + numspec1;
  attrib charrec format = $CHAR.;
  charrec = numrecords;
  charrec = compress(charrec);
  call symput('nrec',charrec);
  do m=1 to numrecords;
    if m = 1 then input +2 mult ibr1. (junk1-junk6) (ibr1.) +24;
    else input +1 mult ibr1. (junk1-junk6) (ibr1.) +24;
    if mult < 0 then mult = mult + 256;
    if mult >= 128 then mult = mult - 256;
    mult = 2 ** (mult - 32);
    do n = 1 to numpoints*4;
      input x(n) ibr1.;
    end;
    do i=4 to numpoints*4 by 4;
      do j = 0 to 3 by 1;
        if x(i-j) < 0 then x(i-j) = x(i-j) + 256;
      end;
      sign = 1;
      if x(i) >= 128 then sign = -1;
      k = i / 4;
      y(k) = (  x(i-3)
              + x(i-2)*256
              + x(i-1)*65536
              + x(i)*16777216 );
      if sign < 0 then y(k) = -(4294967296 - y(k));
      y(k) = mult * y(k);
    end;
    compressedpoints = Floor(newnumpoints / &compressionfactor);
    attrib charcompressedpoints format = $CHAR.;
    charcompressedpoints = compressedpoints;
    charcompressedpoints = compress(charcompressedpoints);
    call symput('comppoints',charcompressedpoints);
    output;
  end;
  stop;
  format y1-y&maxs f13.10;
  keep y1-y&maxs numpoints numrecords;
  run;

  data vspecdata;
  array x(&maxb) x1-x&maxb;
  array y(&maxs) y1-y&maxs;
  infile "&valspecfile" recfm=n;
  input @2 spcversion ibr1.;
  if spcversion < 0 then spcversion = spcversion + 256;
  if spcversion ^= 75 then abort;
  input @5 numpoints1 ibr1. numpoints2 ibr1. @24 +1 numspec1 ibr1. numspec2 ibr1. @512;
  if numpoints1 < 0 then numpoints1 = numpoints1 + 256;
  if numpoints2 < 0 then numpoints2 = numpoints2 + 256;
  numpoints = numpoints2*256 + numpoints1;
  if numspec1 < 0 then numspec1 = numspec1 + 256;
  if numspec2 < 0 then numspec2 = numspec2 + 256;
  numrecords = numspec2*256 + numspec1;
  attrib charrec format = $CHAR.;
  charrec = numrecords;
  charrec = compress(charrec);
  call symput('vnrec',charrec);
  do m=1 to numrecords;
    if m = 1 then input +2 mult ibr1. (junk1-junk6) (ibr1.) +24;
    else input +1 mult ibr1. (junk1-junk6) (ibr1.) +24;
    if mult < 0 then mult = mult + 256;
    if mult >= 128 then mult = mult - 256;
    mult = 2 ** (mult - 32);
    do n = 1 to numpoints*4;
      input x(n) ibr1.;
    end;
    do i=4 to numpoints*4 by 4;
      do j = 0 to 3 by 1;
        if x(i-j) < 0 then x(i-j) = x(i-j) + 256;
      end;
      sign = 1;
      if x(i) >= 128 then sign = -1;
      k = i / 4;
      y(k) = (  x(i-3)
              + x(i-2)*256
              + x(i-1)*65536
              + x(i)*16777216 );
      if sign < 0 then y(k) = -(4294967296 - y(k));
      y(k) = mult * y(k);
    end;
    output;
  end;
  stop;
  format y1-y&maxs f13.10;
  keep y1-y&maxs numpoints numrecords;
  run;

/* ........................................... 2 end ............................................ */

  title1 "PLS Regression. 1st spectral point = &leftspec.. Last spectral point = &rightspec.. Compression factor = &compressionfactor..";
  title2 "specfile: &specfile..";
  title3 "valspecfile: &valspecfile..";
  title4 "chemfile: &chemfile..";
  title5 "valchemfile: &valchemfile..";

%MEND getspec;


/* Macro that reads an ASCII file that contains chemical information in a single column */

%MACRO getchem;

/* ........................................... 3 start .......................................... */

  data chemdata;
  infile "&chemfile";
  input chem1;
  run;

  proc means noprint data = chemdata;
  var chem1;
  output out = chem1var var = v_chem1;
  quit;

  data chem1var2;
  set chem1var;
  mergevarb = 'a';
  run;

  data before_msc_with_chemical;
  merge specdata chemdata;
  mergevariable = 'a';
  run;
  
  data shrink_by_averaging;
  set before_msc_with_chemical;
  array x(&leftspec:&rightspec) y&leftspec-y&rightspec;
  array xreduced(1:&comppoints) z1-z&comppoints;
  do i = (&leftspec + &compressionfactor - 1) to &rightspec by &compressionfactor;
    sumsofar = 0;
    do j = 1 to &compressionfactor;
      sumsofar = sumsofar + x(i - j + 1);
    end;
    aveval = sumsofar/&compressionfactor;
    xreduced((i-&leftspec+1)/&compressionfactor) = aveval;
  end;
  drop y1-y&maxs sumsofar aveval i j;
  run;

  data compress_spectra_with_chemical;
  set shrink_by_averaging;
  rename z1-z&comppoints = y1-y&comppoints;
  run;

  data vchemdata;
  infile "&valchemfile";
  input chem1;
  run;

  data vbefore_msc_with_chemical;
  merge vspecdata vchemdata;
  mergevariable = 'a';
  run;

  data vshrink_by_averaging;
  set vbefore_msc_with_chemical;
  array x(&leftspec:&rightspec) y&leftspec-y&rightspec;
  array xreduced(1:&comppoints) z1-z&comppoints;
  do i = (&leftspec + &compressionfactor - 1) to &rightspec by &compressionfactor;
    sumsofar = 0;
    do j = 1 to &compressionfactor;
      sumsofar = sumsofar + x(i - j + 1);
    end;
    aveval = sumsofar/&compressionfactor;
    xreduced((i-&leftspec+1)/&compressionfactor) = aveval;
  end;
  drop y1-y&maxs sumsofar aveval i j;
  run;

  data vcompress_spectra_with_chemical;
  set vshrink_by_averaging;
  rename z1-z&comppoints = y1-y&comppoints;
  run;
  
  data _null_;
  call symput('leftspec','1');
  call symput('rightspec','&comppoints');
  run;
 
  proc datasets;
  delete before_msc_with_chemical shrink_by_averaging specdata chemdata
  vbefore_msc_with_chemical vshrink_by_averaging vspecdata vchemdata;
  run;
  quit;

/* ........................................... 3 end ............................................ */

%MEND getchem;


/* Macro that reads Savitzky-Golay tables (with corrections) */

%MACRO getSG;

/* ........................................... 4 start .......................................... */

  data savgol;
  infile "&SGfile";
  do table = 1 to &numtables;
    do row = 1 to 14;
      input c1-c11;
      output;
    end;
  end;
  run;

/* ........................................... 4 end ............................................ */

%MEND getSG;


/* Multiplicative Scatter Correction Application Sections */

%MACRO MSCnoSGloop;

  %do k = &pathbegin %to &pathend;

/* ........................................... 5 start .......................................... */

    %if &k = 0 %then %do;

      data path_apply;
      set compress_spectra_with_chemical;
      keep y&leftspec-y&rightspec chem1;
      run;

      data vpath_apply;
      set vcompress_spectra_with_chemical;
      keep y&leftspec-y&rightspec chem1;
      run;

    %end;

    %else %if &k = 1 %then %do;

      proc transpose data=compress_spectra_with_chemical out=t_compress_spec_with_chemical;
      var y&leftspec-y&rightspec;
      quit;

      data t_compress_spec_with_chemicala;
      set t_compress_spec_with_chemical;
      keep col1-col&nrec;
      if col1 = . then delete;
      run;

      proc means data = compress_spectra_with_chemical noprint;
      var y&leftspec-y&rightspec;
      output out=mspectrum mean = my&leftspec-my&rightspec;
      quit;

      proc transpose data=mspectrum out=t_mspectrum;
      var my&leftspec-my&rightspec;
      quit;

      data t_mspectruma;
      set t_mspectrum;
      meanspec = col1;
      keep meanspec;
      if meanspec = . then delete;
      run;

      data t_combin;
      merge t_compress_spec_with_chemicala t_mspectruma;
      run;

      proc reg data = t_combin outest=regest noprint;
      model col1-col&nrec = meanspec;
      quit;

      data path_apply;
      merge compress_spectra_with_chemical regest;
      array x(&leftspec:&rightspec) y&leftspec-y&rightspec;
      do i = &leftspec to &rightspec;
      x(i) = (x(i) - intercept)/meanspec;
      end;
      keep y&leftspec-y&rightspec chem1;
      run;

      proc transpose data=vcompress_spectra_with_chemical out=vt_compress_spec_with_chemical;
      var y&leftspec-y&rightspec;
      quit;

      data vt_compress_spec_with_chemicala;
      set vt_compress_spec_with_chemical;
      keep col1-col&vnrec;
      if col1 = . then delete;
      run;

      data vt_combin;
      merge vt_compress_spec_with_chemicala t_mspectruma;
      run;

      proc reg data = vt_combin outest=vregest noprint;
      model col1-col&vnrec = meanspec;
      quit;

      data vpath_apply;
      merge vcompress_spectra_with_chemical vregest;
      array x(&leftspec:&rightspec) y&leftspec-y&rightspec;
      do i = &leftspec to &rightspec;
      x(i) = (x(i) - intercept)/meanspec;
      end;
      keep y&leftspec-y&rightspec chem1;
      run;

    %end;

/* ........................................... 5 end ............................................ */

    %do m = &varscalebegin %to &varscaleend;

/* ........................................... 6 start .......................................... */

      title6 "MSC without SG section. MSC_loop = &k, varscale_loop = &m..";
      title7 "Program: &progname..  Output file: &SASstyleoutfile..";

      ods listing close;
      %if &m = 0 %then %do;

        proc pls data = path_apply noscale cv = one nfac = &maxfac;
        model chem1 = y&leftspec-y&rightspec;
        output out = plsout predicted = p_chem1;

      %end;

      %else %if &m = 1 %then %do;

        proc pls data = path_apply cv = one nfac = &maxfac;
        model chem1 = y&leftspec-y&rightspec;
        output out = plsout predicted = p_chem1;

      %end;

      ods output ResidualSummary=PRESS_results;
      ods output CVResults=crossval_results;

/* ........................................... 6 end ............................................ */

/* ........................................... 7 start .......................................... */

      data minerror1;
      set crossval_results;
      if _n_ = 1;
      rename nValue1=rootmeanPRESS;
      run;

      data minerror2;
      set crossval_results;
      if _n_ = 2;
      rename nValue1=optimal_factors;
      run;

      data minerror;
      merge minerror1 minerror2;
      run;

      title8 "The R-squared of the optimal model.";

      proc reg data = plsout corr;
      model p_chem1 = chem1;
      ods output Corr=REG_results;

      data getR;
      set REG_results;
      if _n_ = 2;
      R_sq = chem1 * chem1;
      run;

      data onelineout;
      %if &m = 0 %then %do;
        merge minerror getR;
      %end;
      %else %if &m = 1 %then %do;
        merge minerror getR chem1var;
        rootmeanPRESS = rootmeanPRESS * SQRT(v_chem1);
      %end;
      file "&SASstyleoutfile" mod;
      put @1 "SG_Table_= 0" @17 "SG_Window_= 1" @34 "NoMSC(0)_or_MSC(1)_= &k" @59 "Noscale(0)_or_Variance_Scale(1)_= &m"
      @97 "MinPRESS_factors_=" @116 optimal_factors 2.0 @121 "RootmeanPRESS_=" @137 rootmeanPRESS best9. @149
      "R_sq_=" @156 R_sq 6.4;
      run;

      title8 "Summary of SAS Cross Validation Model.";

      proc print data = onelineout;
      var Optimal_factors rootmeanPRESS R_sq;
      quit;

      proc means noprint data=PRESS_results;
      var RootMeanPRESS;
      output out = getthemin min = minRootMeanPRESS;
      quit;

      data minpressvalue;
      set getthemin;
      minPRESSValue = minRootMeanPRESS * minRootMeanPRESS * (&nrec - 1);
      mergevar = 'a';
      keep minPRESSValue mergevar;
      run;

      data addcommonvar;
      set PRESS_results;
      PRESSValue = RootMeanPRESS * RootMeanPRESS * (&nrec - 1);
      mergevar = 'a';
      if _n_ = 1 then delete;
      keep NumberOfFactors PRESSValue mergevar RootMeanPRESS;
      run;

      data gramsliketable;
      merge addcommonvar minpressvalue;
      by mergevar;
      F_Ratio = PRESSValue / minPRESSValue;
      F_Test = CDF('F',F_Ratio,(&nrec-1),(&nrec-1));
      minrec = 0;
      if F_Ratio = 1.0 then minrec = _n_;
      mergevarb = 'a';
      run;

      proc means noprint data = gramsliketable;
      var minrec;
      output out = theminrec max = minfac;
      quit;

      data addamergevar;
      set theminrec;
      mergevarb = 'a';
      run;

/* ........................................... 7 end ............................................ */

/* ........................................... 8 start .......................................... */

      %if &m = 0 %then %do;

        data gramstable;
        merge gramsliketable addamergevar;
        by mergevarb;
        if _n_ > minfac then do;
          F_Ratio = .;
          F_Test = .;
        end;
        run;
      %end;

      %else %if &m = 1 %then %do;

        data pregramstable;
        merge gramsliketable addamergevar;
        by mergevarb;
        if _n_ > minfac then do;
          F_Ratio = .;
          F_Test = .;
        end;
        run;

        data gramstable;
        merge pregramstable chem1var2;
        by mergevarb;
        PRESSValue = PRESSValue * v_chem1;
        run;

      %end;

/* ........................................... 8 end ............................................ */

/* ........................................... 9 start .......................................... */

      title7 "Program: &progname..  Output file: &Gramsstyleoutfile..";

      proc print noobs data = gramstable;
      var NumberOfFactors PRESSValue F_Ratio F_Test;
      quit;

      data trimgrams;
      set gramstable;
      if F_Test > 0.75 then delete;
      run;

      data gramsoptimal;
      set trimgrams;
      if _n_ > 1 then delete;
      run;

      data _null_;
      set gramsoptimal;
      attrib charoptfac format = $CHAR.;
      charoptfac = NumberOfFactors;
      charoptfac = compress(charoptfac);
      call symput('optfac',charoptfac);
      run;

      %if &m = 0 %then %do;

        proc pls data = path_apply noscale nfac=&optfac;
        model chem1 = y&leftspec-y&rightspec / solution;
        output out = plsout predicted = p_chem1;

      %end;

      %else %if &m = 1 %then %do;

        proc pls data = path_apply nfac=&optfac;
        model chem1 = y&leftspec-y&rightspec / solution;
        output out = plsout predicted = p_chem1;

      %end;

      ods output ParameterEstimates=bvector;

/* ........................................... 9 end ............................................ */

/* .......................................... 10 start .......................................... */

      proc reg data = plsout corr;
      model p_chem1 = chem1;
      ods output Corr=REG_results;

      data getR;
      set REG_results;
      if _n_ = 2;
      R_sq = chem1 * chem1;
      run;

      data bvectora;
      set bvector;
      If _n_ = 1 then do;
        attrib charbzero format = $CHAR.;
        charbzero = chem1;
        charbzero = compress(charbzero);
        call symput('bzero',charbzero);
        delete;
      end;
      run;

      proc transpose data=bvectora out=t_bvectora;
      var chem1;
      id RowName;
      quit;

      data t_bvectorb;
      set t_bvectora;
      _TYPE_ = 'SCORE';
      run;

      proc score data=vpath_apply score=t_bvectorb out=valout;
      var y&leftspec-y&rightspec;
      quit;

      data valouta;
      set valout;
      predval = chem12 + &bzero;
      differ = predval - chem1;
      keep predval chem1 differ;
      run;

      proc means noprint data=valouta;
      var differ;
      output out = vstats mean = vBias std=vSEP uss=vUSS n=numsam;
      quit;

      proc reg data = valouta corr;
      model predval = chem1;
      ods output Corr=vREG_results;

      data vgetR;
      set vREG_results;
      if _n_ = 2;
      vr_sq = chem1 * chem1;
      run;

      data valstats;
      merge vstats vgetR;
      vRMSD = sqrt(vUSS/numsam);
      keep vBias vSEP vr_sq vRMSD;
      run;

      data onelineout;
      %if &m = 0 %then %do;
        merge gramsoptimal getR valstats;
        RootMeanPRESS = RootMeanPRESS * SQRT((&nrec-1)/&nrec);
      %end;
      %else %if &m = 1 %then %do;
        merge gramsoptimal getR chem1var valstats;
        RootMeanPRESS = RootMeanPRESS * SQRT((&nrec-1)/&nrec) * SQRT(v_chem1);
      %end;
      file "&Gramsstyleoutfile" mod lrecl = 332;
      rename RootMeanPRESS = RMSD;
      put @1 "SG_Table_= 0" @17 "SG_Window_= 1" @34 "NoMSC(0)_or_MSC(1)_= &k" @59 "Noscale(0)_or_Variance_Scale(1)_= &m"
      @97 "Facs_at_min_error_=" @117 minfac 2.0 @122 "Grams_factors_=" @138 NumberOfFactors 2.0 @143 "PRESS_Value_=" @157
      PRESSValue best10. @170 "F_Ratio_=" @180 F_Ratio best8. @191 "F_Test_=" @200 F_Test best7. @210
      "RMSD_=" @217 RootMeanPRESS best11. @231 "R_sq_=" @238 R_sq 6.4 @247 "vBias_=" @255 vBias best8.
      @266 "vSEP_=" @273 vSEP best11. @287 "vr_sq_=" @295 vr_sq 6.4 @304 "vRMSD_=" @312 vRMSD best11.;
      run;

      title8 "Summary of Grams-Like Cross Validation and Validation.";

      ods listing;
      proc print noobs data = onelineout;
      var NumberOfFactors minfac PRESSValue RMSD F_Ratio F_Test R_sq vBias vSEP vr_sq;
      quit;
      
/* .......................................... 10 end ............................................ */
      
    %end; /* m, variance scaling loop */

  %end; /* k, pathlength or scatter loop */

/* .......................................... 11 start .......................................... */

  %if &pathbegin = 1 | &pathend = 1 %then %do;

    proc datasets;
    delete
      t_compress_spec_with_chemical t_compress_spec_with_chemicala mspectrum
      t_mspectrum t_mspectruma t_combin regest
      path_apply plsout PRESS_results crossval_results minerror1 minerror2
      minerror REG_results getR onelineout getthemin minpressvalue addcommonvar
      gramsliketable theminrec addamergevar gramstable trimgrams gramsoptimal
      vt_compress_spec_with_chemical vt_compress_spec_with_chemicala
      vt_combin vregest
      vpath_apply
      bvector bvectora t_bvectora t_bvectorb valout valouta
      vstats vREG_results vgetR valstats;
    run; quit;

  %end;

  %else %do;

    proc datasets;
    delete
      path_apply plsout PRESS_results crossval_results minerror1 minerror2
      minerror REG_results getR onelineout getthemin minpressvalue addcommonvar
      gramsliketable theminrec addamergevar gramstable trimgrams gramsoptimal
      vpath_apply
      bvector bvectora t_bvectora t_bvectorb valout valouta
      vstats vREG_results vgetR valstats;
    run; quit;

  %end;

  %if &varscalebegin = 1 | &varscaleend = 1 %then %do;

    proc datasets;
    delete
      pregramstable;
    run; quit;

  %end;

/* .......................................... 11 end ............................................ */

%MEND MSCnoSGloop;


%MACRO SGloop;

  %do i=&tablebegin %to &tableend;

/* .......................................... 12 start .......................................... */

    data chosentable;
    set savgol;
    if &i = 1 then convolutiontable = &firstSGtable;
    else if &i = 2 then convolutiontable = &secondSGtable;
    else convolutiontable = &thirdSGtable;
    if table ^= convolutiontable then delete;
    run;

    proc transpose data=chosentable out=t_chosentable;
    var c1-c11;
    quit;

/* .......................................... 12 end ............................................ */

    %do j=&windowbegin %to &windowend %by 2;
      %if
      (
      ^(&i = 1 and (&firstSGtable = 2 or &firstSGtable = 5 or &firstSGtable = 7) and &j = 5)
      and
      ^(&i = 2 and (&secondSGtable = 2 or &secondSGtable = 5 or &secondSGtable = 7) and &j = 5)
      and
      ^(&i = 3 and (&thirdSGtable = 2 or &thirdSGtable = 5 or &thirdSGtable = 7) and &j = 5)
      )
      %then %do; /* normal route, except when tables 2, 5, or 7 are chosen with 5 convolution pts */

/* .......................................... 13 start .......................................... */

        data convolute;
        set t_chosentable;
        if &i = 1 then table = &firstSGtable;
        else if &i = 2 then table = &secondSGtable;
        else table = &thirdSGtable;
        SG_pts = &j;
        rename col1=r1 col2=r2 col3=r3 col4=r4 col5=r5 col6=r6 col7=r7 col8=r8 col9=r9
               col10=r10 col11=r11 col12=r12 col13=r13 col14=r14;
        if table = 3 or table = 4 or table = 5 or table = 9 then sign = -1; else sign = 1;
        mergevariable = 'a';
        keep sign SG_pts table col1-col14 mergevariable;
        if _n_ ^= 13.5 - SG_pts/2 then delete;
        run;

        data deriv;
        merge compress_spectra_with_chemical convolute;
        by mergevariable;
        array y(&leftspec:&rightspec) y&leftspec-y&rightspec;
        array r(14) r1-r14;
        array z(&leftspec:&rightspec) z&leftspec-z&rightspec;
        newleft = (SG_pts+1)/2 + (&leftspec-1);
        newright = &rightspec - (SG_pts-1)/2;
        attrib charleft format = $CHAR.;
        charleft = newleft;
        charleft = compress(charleft);
        call symput('lefttrim',charleft);
        attrib charright format = $CHAR.;
        charright = newright;
        charright = compress(charright);
        call symput('righttrim',charright);
        do a = newleft to newright;
          sum = 0.0;
          do aa = (14 - (SG_pts+1)/2) to 13;
            sum = sum +        r(aa) * y(a-(13-aa));
          end;
          do aa = (14 - (SG_pts+1)/2) to 12;
            sum = sum + sign * r(aa) * y(a+(13-aa));
          end;
          z(a) = sum / r(14);
        end;
        keep z&leftspec-z&rightspec chem1 SG_pts table newleft newright numpoints numrecords mergevariable;
        run;

        data vderiv;
        merge vcompress_spectra_with_chemical convolute;
        by mergevariable;
        array y(&leftspec:&rightspec) y&leftspec-y&rightspec;
        array r(14) r1-r14;
        array z(&leftspec:&rightspec) z&leftspec-z&rightspec;
        newleft = (SG_pts+1)/2 + (&leftspec-1);
        newright = &rightspec - (SG_pts-1)/2;
        do a = newleft to newright;
          sum = 0.0;
          do aa = (14 - (SG_pts+1)/2) to 13;
            sum = sum +        r(aa) * y(a-(13-aa));
          end;
          do aa = (14 - (SG_pts+1)/2) to 12;
            sum = sum + sign * r(aa) * y(a+(13-aa));
          end;
          z(a) = sum / r(14);
        end;
        keep z&leftspec-z&rightspec chem1 SG_pts table newleft newright numpoints numrecords mergevariable;
        run;

/* .......................................... 13 end ............................................ */

        %do k = &pathbegin %to &pathend;

/* .......................................... 14 start .......................................... */

          %if &k = 0 %then %do;

            data path_apply;
            set deriv;
            keep z&lefttrim-z&righttrim chem1 SG_pts table newleft newright numpoints;
            run;

            data vpath_apply;
            set vderiv;
            keep z&lefttrim-z&righttrim chem1 SG_pts table newleft newright numpoints;
            run;

          %end;

          %else %if &k = 1 %then %do;

            proc transpose data=deriv out=t_deriv;
            var z&lefttrim-z&righttrim;
            quit;

            data t_deriva;
            set t_deriv;
            keep col1-col&nrec;
            if col1 = . then delete;
            run;

            proc means data = deriv noprint;
            var z&lefttrim-z&righttrim;
            output out=mspectrum mean = mz&lefttrim-mz&righttrim;
            quit;

            proc transpose data=mspectrum out=t_mspectrum;
            var mz&lefttrim-mz&righttrim;
            quit;

            data t_mspectruma;
            set t_mspectrum;
            meanspec = col1;
            keep meanspec;
            if meanspec = . then delete;
            run;

            data t_combin;
            merge t_deriva t_mspectruma;
            run;

            proc reg data = t_combin outest=regest noprint;
            model col1-col&nrec = meanspec;
            quit;

            data path_apply;
            merge deriv regest;
            array x(&lefttrim:&righttrim) z&lefttrim-z&righttrim;
            do i = &lefttrim to &righttrim;
              x(i) = (x(i) - intercept)/meanspec;
            end;
            keep z&lefttrim-z&righttrim chem1 SG_pts table newleft newright numpoints;
            run;

            proc transpose data=vderiv out=vt_deriv;
            var z&lefttrim-z&righttrim;
            quit;

            data vt_deriva;
            set vt_deriv;
            keep col1-col&vnrec;
            if col1 = . then delete;
            run;

            data vt_combin;
            merge vt_deriva t_mspectruma;
            run;

            proc reg data = vt_combin outest=vregest noprint;
            model col1-col&vnrec = meanspec;
            quit;

            data vpath_apply;
            merge vderiv vregest;
            array x(&lefttrim:&righttrim) z&lefttrim-z&righttrim;
            do i = &lefttrim to &righttrim;
              x(i) = (x(i) - intercept)/meanspec;
            end;
            keep z&lefttrim-z&righttrim chem1 SG_pts table newleft newright numpoints;
            run;

          %end;

/* .......................................... 14 end ............................................ */

          %do m = &varscalebegin %to &varscaleend;

/* .......................................... 15 start .......................................... */

            title6 "MSC with SG section. table_loop = &i, convwindow_loop = &j, MSC_loop = &k, varscale_loop = &m..";
            title7 "Program: &progname..  Output file: &SASstyleoutfile..";

            ods listing close;
            %if &m = 0 %then %do;

              proc pls data = path_apply noscale cv = one nfac = &maxfac;
              model chem1 = z&lefttrim-z&righttrim;
              output out = plsout predicted = p_chem1;

            %end;

            %else %if &m = 1 %then %do;

              proc pls data = path_apply cv = one nfac = &maxfac;
              model chem1 = z&lefttrim-z&righttrim;
              output out = plsout predicted = p_chem1;

            %end;

            ods output ResidualSummary=PRESS_results;
            ods output CVResults=crossval_results;

/* .......................................... 15 end ............................................ */

/* .......................................... 16 start .......................................... */

            data minerror1;
            set crossval_results;
            if _n_ = 1;
            rename nValue1=rootmeanPRESS;
            run;

            data minerror2;
            set crossval_results;
            if _n_ = 2;
            rename nValue1=optimal_factors;
            run;

            data minerror;
            merge minerror1 minerror2;
            run;

            title8 "The R-squared of the optimal model.";

            proc reg data = plsout corr;
            model p_chem1 = chem1;
            ods output Corr=REG_results;

            data getR;
            set REG_results;
            if _n_ = 2;
            R_sq = chem1 * chem1;
            run;

            data onelineout;
            %if &m = 0 %then %do;
              merge minerror getR;
            %end;
            %else %if &m = 1 %then %do;
              merge minerror getR chem1var;
              rootmeanPRESS = rootmeanPRESS * SQRT(v_chem1);
            %end;
            file "&SASstyleoutfile" mod;
            %if       &i = 1 %then %do;
              put @1 "SG_Table_= &firstSGtable" @;
            %end;
            %else %if &i = 2 %then %do;
              put @1 "SG_Table_= &secondSGtable" @;
            %end;
            %else %if &i = 3 %then %do;
              put @1 "SG_Table_= &thirdSGtable" @;
            %end;
            put @17 "SG_Window_= &j" @34 "NoMSC(0)_or_MSC(1)_= &k" @59
            "Noscale(0)_or_Variance_Scale(1)_= &m" @97 "MinPRESS_factors_=" @116 optimal_factors 2.0 @121
            "RootmeanPRESS_=" @137 rootmeanPRESS best9. @149 "R_sq_=" @156 R_sq 6.4;
            run;

            title8 "Summary of SAS Cross Validation Model.";

            proc print data = onelineout;
            var Optimal_factors rootmeanPRESS R_sq;
            quit;

            proc means noprint data=PRESS_results;
            var RootMeanPRESS;
            output out = getthemin min = minRootMeanPRESS;
            quit;

            data minpressvalue;
            set getthemin;
            minPRESSValue = minRootMeanPRESS * minRootMeanPRESS * (&nrec - 1);
            mergevar = 'a';
            keep minPRESSValue mergevar;
            run;

            data addcommonvar;
            set PRESS_results;
            PRESSValue = RootMeanPRESS * RootMeanPRESS * (&nrec - 1);
            mergevar = 'a';
            if _n_ = 1 then delete;
            keep NumberOfFactors PRESSValue mergevar RootMeanPRESS;
            run;

            data gramsliketable;
            merge addcommonvar minpressvalue;
            by mergevar;
            F_Ratio = PRESSValue / minPRESSValue;
            F_Test = CDF('F',F_Ratio,(&nrec-1),(&nrec-1));
            minrec = 0;
            if F_Ratio = 1.0 then minrec = _n_;
            mergevarb = 'a';
            run;

            proc means noprint data = gramsliketable;
            var minrec;
            output out = theminrec max = minfac;
            quit;

            data addamergevar;
            set theminrec;
            mergevarb = 'a';
            run;

/* .......................................... 16 end ............................................ */

/* .......................................... 17 start .......................................... */

           %if &m = 0 %then %do;

              data gramstable;
              merge gramsliketable addamergevar;
              by mergevarb;
              if _n_ > minfac then do;
                F_Ratio = .;
                F_Test = .;
              end;
              run;
            %end;

            %else %if &m = 1 %then %do;

              data pregramstable;
              merge gramsliketable addamergevar;
              by mergevarb;
              if _n_ > minfac then do;
                F_Ratio = .;
                F_Test = .;
              end;
              run;

              data gramstable;
              merge pregramstable chem1var2;
              by mergevarb;
              PRESSValue = PRESSValue * v_chem1;
              run;

            %end;

/* .......................................... 17 end ............................................ */

/* .......................................... 18 start .......................................... */

            title7 "Program: &progname..  Output file: &Gramsstyleoutfile..";

            proc print noobs data = gramstable;
            var NumberOfFactors PRESSValue F_Ratio F_Test;
            quit;

            data trimgrams;
            set gramstable;
            if F_Test > 0.75 then delete;
            run;

            data gramsoptimal;
            set trimgrams;
            if _n_ > 1 then delete;
            run;

            data _null_;
            set gramsoptimal;
            attrib charoptfac format = $CHAR.;
            charoptfac = NumberOfFactors;
            charoptfac = compress(charoptfac);
            call symput('optfac',charoptfac);
            run;

            %if &m = 0 %then %do;

              proc pls data = path_apply noscale nfac=&optfac;
              model chem1 = z&lefttrim-z&righttrim / solution;
              output out = plsout predicted = p_chem1;

            %end;

            %else %if &m = 1 %then %do;

              proc pls data = path_apply nfac=&optfac;
              model chem1 = z&lefttrim-z&righttrim / solution;
              output out = plsout predicted = p_chem1;

            %end;
            ods output ParameterEstimates=bvector;

/* .......................................... 18 end ............................................ */

/* .......................................... 19 start .......................................... */

            proc reg data = plsout corr;
            model p_chem1 = chem1;
            ods output Corr=REG_results;

            data getR;
            set REG_results;
            if _n_ = 2;
            R_sq = chem1 * chem1;
            run;
            
            data bvectora;
            set bvector;
            If _n_ = 1 then do;
              attrib charbzero format = $CHAR.;
              charbzero = chem1;
              charbzero = compress(charbzero);
              call symput('bzero',charbzero);
              delete;
            end;
            run;

            proc transpose data=bvectora out=t_bvectora;
            var chem1;
            id RowName;
            quit;

            data t_bvectorb;
            set t_bvectora;
            _TYPE_ = 'SCORE';
            run;

            proc score data=vpath_apply score=t_bvectorb out=valout;
            var z&lefttrim-z&righttrim;
            quit;

            data valouta;
            set valout;
            predval = chem12 + &bzero;
            differ = predval - chem1;
            keep predval chem1 differ;
            run;

            proc means noprint data=valouta;
            var differ;
            output out = vstats mean = vBias std=vSEP uss=vUSS n=numsam;
            quit;

            proc reg data = valouta corr;
            model predval = chem1;
            ods output Corr=vREG_results;

            data vgetR;
            set vREG_results;
            if _n_ = 2;
            vr_sq = chem1 * chem1;
            run;

            data valstats;
            merge vstats vgetR;
            vRMSD = sqrt(vUSS/numsam);
            keep vBias vSEP vr_sq vRMSD;
            run;

            data onelineout;
            %if &m = 0 %then %do;
              merge gramsoptimal getR valstats;
              RootMeanPRESS = RootMeanPRESS * SQRT((&nrec-1)/&nrec);
            %end;
            %else %if &m = 1 %then %do;
              merge gramsoptimal getR chem1var valstats;
              RootMeanPRESS = RootMeanPRESS * SQRT((&nrec-1)/&nrec) * SQRT(v_chem1);
            %end;
            file "&Gramsstyleoutfile" mod lrecl = 332;
            rename RootMeanPRESS = RMSD;
            %if      &i = 1 %then %do;
              put @1 "SG_Table_= &firstSGtable" @;
            %end;
            %else %if &i = 2 %then %do;
              put @1 "SG_Table_= &secondSGtable" @;
            %end;
            %else %if &i = 3 %then %do;
              put @1 "SG_Table_= &thirdSGtable" @;
            %end;
            put @17 "SG_Window_= &j" @34 "NoMSC(0)_or_MSC(1)_= &k" @59
            "Noscale(0)_or_Variance_Scale(1)_= &m" @97 "Facs_at_min_error_=" @117 minfac 2.0 @122
            "Grams_factors_=" @138 NumberOfFactors 2.0 @143 "PRESS_Value_=" @157 PRESSValue best10.
            @170 "F_Ratio_=" @180 F_Ratio best8. @191 "F_Test_=" @200 F_Test best7. @210
            "RMSD_=" @217 RootMeanPRESS best11. @231 "R_sq_=" @238 R_sq 6.4 @247 "vBias_=" @255 vBias best8.
            @266 "vSEP_=" @273 vSEP best11. @287 "vr_sq_=" @295 vr_sq 6.4 @304 "vRMSD_=" @312 vRMSD best11.;
            run;

            title8 "Summary of Grams-Like Cross Validation and Validation.";

            ods listing;
            proc print noobs data = onelineout;
            var NumberOfFactors minfac PRESSValue RMSD F_Ratio F_Test R_sq vBias vSEP vr_sq vRMSD;
            quit;

/* .......................................... 19 end ............................................ */

          %end; /* m, variance scaling loop */

        %end; /* k, pathlength or scatter loop */

      %end; /* the normal route, except when table is 2, 5, or 7 and conv. pts = 5 */

    %end; /* j, SG convolution window loop */

  %end; /* i, SG table loop */

/* .......................................... 20 start .......................................... */

  %if &pathbegin = 1 | &pathend = 1 %then %do;

    proc datasets;
    delete
      chosentable t_chosentable convolute deriv
      t_deriv t_deriva mspectrum t_mspectrum t_mspectruma t_combin regest
      path_apply plsout PRESS_results crossval_results minerror1 minerror2 minerror
      REG_results getR onelineout getthemin minpressvalue addcommonvar gramsliketable theminrec
      addamergevar gramstable trimgrams gramsoptimal
      vderiv vpathapply
      vt_deriv vt_deriva vt_combin vregest
      bvector bvectora t_bvectora t_bvectorb valout valouta
      vstats vREG_results vgetR valstats;
    run; quit;

  %end;

  %else %do;

    proc datasets;
    delete
      chosentable t_chosentable convolute deriv
      path_apply plsout PRESS_results crossval_results minerror1 minerror2 minerror
      REG_results getR onelineout getthemin minpressvalue addcommonvar gramsliketable theminrec
      addamergevar gramstable trimgrams gramsoptimal
      vderiv vpathapply
      bvector bvectora t_bvectora t_bvectorb valout valouta
      vstats vREG_results vgetR valstats;
    run; quit;

  %end;

  %if &varscalebegin = 1 | &varscaleend = 1 %then %do;

    proc datasets;
    delete
      pregramstable;
    run; quit;

  %end;

/* .......................................... 20 end ............................................ */

%MEND SGloop;


/* Standard Normal Variate Application Sections */

%MACRO SNVnoSGloop;

/* .......................................... 21 start .......................................... */

  proc transpose data=compress_spectra_with_chemical out=t_before_msc;
  var y&leftspec-y&rightspec;
  quit;

  data t_before_msca;
  set t_before_msc;
  keep col1-col&nrec;
  if col1 = . then delete;
  run;

  proc means data = t_before_msca noprint;
  var col1-col&nrec;
  output out=mspectrum mean = mcol1-mcol&nrec std = scol1-scol&nrec;
  quit;

  proc transpose data=mspectrum out=t_mspectrum_mean;
  var mcol1-mcol&nrec;
  quit;

  data addmean;
  merge compress_spectra_with_chemical t_mspectrum_mean;
  rename col1=specmean;
  run;

  proc transpose data=mspectrum out=t_mspectrum_std;
  var scol1-scol&nrec;
  quit;

  data SNV_apply;
  merge addmean t_mspectrum_std;
  array x(&leftspec:&rightspec) y&leftspec-y&rightspec;
  do p = &leftspec to &rightspec;
    x(p) = (x(p) - specmean)/col1;
  end;
  rename col1=specstd;
  keep y&leftspec-y&rightspec chem1 numpoints numrecords specmean col1;
  run;

  proc transpose data=vcompress_spectra_with_chemical out=vt_before_msc;
  var y&leftspec-y&rightspec;
  quit;

  data vt_before_msca;
  set vt_before_msc;
  keep col1-col&vnrec;
  if col1 = . then delete;
  run;

  proc means data = vt_before_msca noprint;
  var col1-col&vnrec;
  output out=vmspectrum mean = mcol1-mcol&vnrec std = scol1-scol&vnrec;
  quit;

  proc transpose data=vmspectrum out=vt_mspectrum_mean;
  var mcol1-mcol&vnrec;
  quit;

  data vaddmean;
  merge vcompress_spectra_with_chemical vt_mspectrum_mean;
  rename col1=specmean;
  run;

  proc transpose data=vmspectrum out=vt_mspectrum_std;
  var scol1-scol&vnrec;
  quit;

  data vSNV_apply;
  merge vaddmean vt_mspectrum_std;
  array x(&leftspec:&rightspec) y&leftspec-y&rightspec;
  do p = &leftspec to &rightspec;
    x(p) = (x(p) - specmean)/col1;
  end;
  rename col1=specstd;
  keep y&leftspec-y&rightspec chem1 numpoints numrecords specmean col1;
  run;
  

/* .......................................... 21 end ............................................ */

  %do i=&detrendbegin %to &detrendend;

/* .......................................... 22 start .......................................... */

    %if &i = 0 %then %do;

      data afterSNV;
      set SNV_apply;
      mergevariable = 'a';
      keep y&leftspec-y&rightspec chem1 numpoints numrecords mergevariable;
      run;

      data vafterSNV;
      set vSNV_apply;
      mergevariable = 'a';
      keep y&leftspec-y&rightspec chem1 numpoints numrecords mergevariable;
      run;

    %end;

    %else %if &i = 1 %then %do;

      proc transpose data=SNV_apply out=t_SNV_apply;
      var y&leftspec-y&rightspec;
      quit;

      data t_SNV_applyb;
      set t_SNV_apply;
      n = _n_;
      n_sq = n * n;
      keep col1-col&nrec n n_sq;
      run;

      proc reg data = t_SNV_applyb noprint;
      model col1-col&nrec = n n_sq;
      output out = detrend_apply  residual = resid1-resid&nrec;
      quit;

      proc transpose data=detrend_apply out=t_detrend_apply;
      var resid1-resid&nrec;
      quit;

      data detrendresid;
      set t_detrend_apply;
      array y(&leftspec:&rightspec) col&leftspec-col&rightspec;
      array z(&leftspec:&rightspec) z&leftspec-z&rightspec;
      do p = &leftspec to &rightspec;
        z(p) = y(p);
      end;
      keep z&leftspec-z&rightspec;
      run;

      data afterSNV;
      merge SNV_apply detrendresid;
      array y(&leftspec:&rightspec) y&leftspec-y&rightspec;
      array z(&leftspec:&rightspec) z&leftspec-z&rightspec;
      do l = &leftspec to &rightspec;
        y(l) = z(l);
      end;
      mergevariable = 'a';
      keep y&leftspec-y&rightspec chem1 numpoints numrecords mergevariable;
      run;

      proc transpose data=vSNV_apply out=vt_SNV_apply;
      var y&leftspec-y&rightspec;
      quit;

      data vt_SNV_applyb;
      set vt_SNV_apply;
      n = _n_;
      n_sq = n * n;
      keep col1-col&vnrec n n_sq;
      run;

      proc reg data = vt_SNV_applyb noprint;
      model col1-col&vnrec = n n_sq;
      output out = vdetrend_apply  residual = resid1-resid&vnrec;
      quit;

      proc transpose data=vdetrend_apply out=vt_detrend_apply;
      var resid1-resid&vnrec;
      quit;

      data vdetrendresid;
      set vt_detrend_apply;
      array y(&leftspec:&rightspec) col&leftspec-col&rightspec;
      array z(&leftspec:&rightspec) z&leftspec-z&rightspec;
      do p = &leftspec to &rightspec;
        z(p) = y(p);
      end;
      keep z&leftspec-z&rightspec;
      run;

      data vafterSNV;
      merge vSNV_apply vdetrendresid;
      array y(&leftspec:&rightspec) y&leftspec-y&rightspec;
      array z(&leftspec:&rightspec) z&leftspec-z&rightspec;
      do l = &leftspec to &rightspec;
        y(l) = z(l);
      end;
      mergevariable = 'a';
      keep y&leftspec-y&rightspec chem1 numpoints numrecords mergevariable;
      run;

    %end;

/* .......................................... 22 end ............................................ */

    %do m = &varscalebegin %to &varscaleend;

/* .......................................... 23 start .......................................... */

      title6 "SNV without SG section. detrend_loop = &i, varscale_loop = &m..";
      title7 "Program: &progname..  Output file: &SASstyleoutfile..";

      ods listing close;
      %if &m = 0 %then %do;

        proc pls data = afterSNV noscale cv = one nfac = &maxfac;
        model chem1 = y&leftspec-y&rightspec;
        output out = plsout predicted = p_chem1;

      %end;

      %else %if &m = 1 %then %do;

        proc pls data = afterSNV cv = one nfac = &maxfac;
        model chem1 = y&leftspec-y&rightspec;
        output out = plsout predicted = p_chem1;

      %end;

      ods output ResidualSummary=PRESS_results;
      ods output CVResults=crossval_results;

/* .......................................... 23 end ............................................ */

/* .......................................... 24 start .......................................... */

      data minerror1;
      set crossval_results;
      if _n_ = 1;
      rename nValue1=rootmeanPRESS;
      run;

      data minerror2;
      set crossval_results;
      if _n_ = 2;
      rename nValue1=optimal_factors;
      run;

      data minerror;
      merge minerror1 minerror2;
      run;

      title8 "The R-squared of the optimal model.";

      proc reg data = plsout corr;
      model p_chem1 = chem1;
      ods output Corr=REG_results;

      data getR;
      set REG_results;
      if _n_ = 2;
      R_sq = chem1 * chem1;
      run;

      data onelineout;
      %if &m = 0 %then %do;
        merge minerror getR;
      %end;
      %else %if &m = 1 %then %do;
        merge minerror getR chem1var;
        rootmeanPRESS = rootmeanPRESS * SQRT(v_chem1);
      %end;
      file "&SASstyleoutfile" mod;
      put @1 "SG_Table_= 0" @17 "SG_Window_= 1" @34 "SNV(0)_or_SNVdt(1)_= &i" @59 "Noscale(0)_or_Variance_Scale(1)_= &m"
      @97 "MinPRESS_factors_=" @116 optimal_factors 2.0 @121 "RootmeanPRESS_=" @137 rootmeanPRESS best9. @149
      "R_sq_=" @156 R_sq 6.4;
      run;

      title8 "Summary of SAS Cross Validation Model, SNV section.";

      proc print data = onelineout;
      var Optimal_factors rootmeanPRESS R_sq;
      quit;

      proc means noprint data=PRESS_results;
      var RootMeanPRESS;
      output out = getthemin min = minRootMeanPRESS;
      quit;

      data minpressvalue;
      set getthemin;
      minPRESSValue = minRootMeanPRESS * minRootMeanPRESS * (&nrec - 1);
      mergevar = 'a';
      keep minPRESSValue mergevar;
      run;

      data addcommonvar;
      set PRESS_results;
      PRESSValue = RootMeanPRESS * RootMeanPRESS * (&nrec - 1);
      mergevar = 'a';
      if _n_ = 1 then delete;
      keep NumberOfFactors PRESSValue mergevar RootMeanPRESS;
      run;

      data gramsliketable;
      merge addcommonvar minpressvalue;
      by mergevar;
      F_Ratio = PRESSValue / minPRESSValue;
      F_Test = CDF('F',F_Ratio,(&nrec-1),(&nrec-1));
      minrec = 0;
      if F_Ratio = 1.0 then minrec = _n_;
      mergevarb = 'a';
      run;

      proc means noprint data = gramsliketable;
      var minrec;
      output out = theminrec max = minfac;
      quit;

      data addamergevar;
      set theminrec;
      mergevarb = 'a';
      run;

/* .......................................... 24 end ............................................ */

/* .......................................... 25 start .......................................... */

      %if &m = 0 %then %do;

        data gramstable;
        merge gramsliketable addamergevar;
        by mergevarb;
        if _n_ > minfac then do;
          F_Ratio = .;
          F_Test = .;
        end;
        run;
      %end;

      %else %if &m = 1 %then %do;

        data pregramstable;
        merge gramsliketable addamergevar;
        by mergevarb;
        if _n_ > minfac then do;
          F_Ratio = .;
          F_Test = .;
        end;
        run;

        data gramstable;
        merge pregramstable chem1var2;
        by mergevarb;
        PRESSValue = PRESSValue * v_chem1;
        run;

      %end;

/* .......................................... 25 end ............................................ */

/* .......................................... 26 start .......................................... */

      title7 "Program: &progname..  Output file: &Gramsstyleoutfile..";

      proc print noobs data = gramstable;
      var NumberOfFactors PRESSValue F_Ratio F_Test;
      quit;

      data trimgrams;
      set gramstable;
      if F_Test > 0.75 then delete;
      run;

      data gramsoptimal;
      set trimgrams;
      if _n_ > 1 then delete;
      run;

      data _null_;
      set gramsoptimal;
      attrib charoptfac format = $CHAR.;
      charoptfac = NumberOfFactors;
      charoptfac = compress(charoptfac);
      call symput('optfac',charoptfac);
      run;

      %if &m = 0 %then %do;

        proc pls data = afterSNV noscale nfac=&optfac;
        model chem1 = y&leftspec-y&rightspec / solution;
        output out = plsout predicted = p_chem1;

      %end;

      %else %if &m = 1 %then %do;

        proc pls data = afterSNV nfac=&optfac;
        model chem1 = y&leftspec-y&rightspec / solution;
        output out = plsout predicted = p_chem1;

      %end;
      
      ods output ParameterEstimates=bvector;

/* .......................................... 26 end ............................................ */

/* .......................................... 27 start .......................................... */

      proc reg data = plsout corr;
      model p_chem1 = chem1;
      ods output Corr=REG_results;

      data getR;
      set REG_results;
      if _n_ = 2;
      R_sq = chem1 * chem1;
      run;

      data bvectora;
      set bvector;
      If _n_ = 1 then do;
        attrib charbzero format = $CHAR.;
        charbzero = chem1;
        charbzero = compress(charbzero);
        call symput('bzero',charbzero);
        delete;
      end;
      run;

      proc transpose data=bvectora out=t_bvectora;
      var chem1;
      id RowName;
      quit;

      data t_bvectorb;
      set t_bvectora;
      _TYPE_ = 'SCORE';
      run;

      proc score data=vafterSNV score=t_bvectorb out=valout;
      var y&leftspec-y&rightspec;
      quit;

      data valouta;
      set valout;
      predval = chem12 + &bzero;
      differ = predval - chem1;
      keep predval chem1 differ;
      run;

      proc means noprint data=valouta;
      var differ;
      output out = vstats mean = vBias std=vSEP uss=vUSS n=numsam;
      quit;

      proc reg data = valouta corr;
      model predval = chem1;
      ods output Corr=vREG_results;

      data vgetR;
      set vREG_results;
      if _n_ = 2;
      vr_sq = chem1 * chem1;
      run;

      data valstats;
      merge vstats vgetR;
      vRMSD = sqrt(vUSS/numsam);
      keep vBias vSEP vr_sq vRMSD;
      run;

      data onelineout;
      %if &m = 0 %then %do;
        merge gramsoptimal getR valstats;
        RootMeanPRESS = RootMeanPRESS * SQRT((&nrec-1)/&nrec);
      %end;
      %else %if &m = 1 %then %do;
        merge gramsoptimal getR chem1var valstats;
        RootMeanPRESS = RootMeanPRESS * SQRT((&nrec-1)/&nrec) * SQRT(v_chem1);
      %end;
      file "&Gramsstyleoutfile" mod lrecl = 325;
      rename RootMeanPRESS = RMSD;
      put @1 "SG_Table_= 0" @17 "SG_Window_= 1" @34 "SNV(0)_or_SNVdt(1)_= &i" @59
      "Noscale(0)_or_Variance_Scale(1)_= &m" @97 "Facs_at_min_error_=" @117 minfac 2.0 @122
      "Grams_factors_=" @138 NumberOfFactors 2.0 @143 "PRESS_Value_=" @157 PRESSValue best10.
      @170 "F_Ratio_=" @180 F_Ratio best8. @191 "F_Test_=" @200 F_Test best7. @210
      "RMSD_=" @217 RootMeanPRESS best11. @231 "R_sq_=" @238 R_sq 6.4 @247 "vBias_=" @255 vBias best8.
      @266 "vSEP_=" @273 vSEP best11. @287 "vr_sq_=" @295 vr_sq 6.4 @304 "vRMSD_=" @312 vRMSD best11.;
      run;

      title8 "Summary of Grams-Like Cross Validation and Validation.";
      ods listing;
      proc print noobs data = onelineout;
      var NumberOfFactors minfac PRESSValue RMSD F_Ratio F_Test R_sq vBias vSEP vr_sq vRMSD;
      quit;

/* .......................................... 27 end ............................................ */

    %end; /* m, variance scaling loop */

  %end; /* i, SNV detrend loop */

/* .......................................... 28 start .......................................... */

  %if &detrendbegin = 1 | &detrendend = 1 %then %do;

    proc datasets;
    delete
      t_before_msc t_before_msca mspectrum t_mspectrum_mean addmean t_mspectrum_std SNV_apply
      t_SNV_apply t_SNV_applyb detrend_apply t_detrend_apply detrendresid
      afterSNV plsout PRESS_results crossval_results minerror1 minerror2 minerror
      REG_results getR onelineout getthemin minpressvalue addcommonvar gramsliketable
      theminrec addamergevar gramstable trimgrams gramsoptimal
      vt_before_msc vt_before_msca vmspectrum vt_mspectrum_mean vaddmean vt_mspectrum_std vSNV_apply
      vt_SNV_apply vt_SNV_applyb vdetrend_apply vt_detrend_apply vdetrendresid
      vafterSNV
      bvector bvectora t_bvectora t_bvectorb valout valouta
      vstats vREG_results vgetR valstats;
    run; quit;

  %end;

  %else %do;

    proc datasets;
    delete
      t_before_msc t_before_msca mspectrum t_mspectrum_mean addmean t_mspectrum_std SNV_apply
      afterSNV plsout PRESS_results crossval_results minerror1 minerror2 minerror
      REG_results getR onelineout getthemin minpressvalue addcommonvar gramsliketable
      theminrec addamergevar gramstable trimgrams gramsoptimal
      vt_before_msc vt_beforemsca vmspectrum vt_mspectrum_mean vaddmean vt_mspectrum_std vSNV_apply
      vafterSNV
      bvector bvectora t_bvectora t_bvectorb valout valouta
      vstats vREG_results vgetR valstats;
    run; quit;

  %end;

  %if &varscalebegin = 1 | &varscaleend = 1 %then %do;

    proc datasets;
    delete
      pregramstable;
    run; quit;

  %end;

/* .......................................... 28 end ............................................ */

%MEND SNVnoSGloop;


%MACRO SNVloop;

/* .......................................... 29 start .......................................... */

  proc transpose data=compress_spectra_with_chemical out=t_before_msc;
  var y&leftspec-y&rightspec;
  quit;

  data t_before_msca;
  set t_before_msc;
  keep col1-col&nrec;
  if col1 = . then delete;
  run;

  proc means data = t_before_msca noprint;
  var col1-col&nrec;
  output out=mspectrum mean = mcol1-mcol&nrec std = scol1-scol&nrec;
  quit;

  proc transpose data=mspectrum out=t_mspectrum_mean;
  var mcol1-mcol&nrec;
  quit;

  data addmean;
  merge compress_spectra_with_chemical t_mspectrum_mean;
  rename col1=specmean;
  run;

  proc transpose data=mspectrum out=t_mspectrum_std;
  var scol1-scol&nrec;
  quit;

  data SNV_apply;
  merge addmean t_mspectrum_std;
  array x(&leftspec:&rightspec) y&leftspec-y&rightspec;
  do p = &leftspec to &rightspec;
    x(p) = (x(p) - specmean)/col1;
  end;
  rename col1=specstd;
  keep y&leftspec-y&rightspec chem1 numpoints numrecords specmean col1;
  run;

  proc transpose data=vcompress_spectra_with_chemical out=vt_before_msc;
  var y&leftspec-y&rightspec;
  quit;

  data vt_before_msca;
  set vt_before_msc;
  keep col1-col&vnrec;
  if col1 = . then delete;
  run;

  proc means data = vt_before_msca noprint;
  var col1-col&vnrec;
  output out=vmspectrum mean = mcol1-mcol&vnrec std = scol1-scol&vnrec;
  quit;

  proc transpose data=vmspectrum out=vt_mspectrum_mean;
  var mcol1-mcol&vnrec;
  quit;

  data vaddmean;
  merge vcompress_spectra_with_chemical vt_mspectrum_mean;
  rename col1=specmean;
  run;

  proc transpose data=vmspectrum out=vt_mspectrum_std;
  var scol1-scol&vnrec;
  quit;

  data vSNV_apply;
  merge vaddmean vt_mspectrum_std;
  array x(&leftspec:&rightspec) y&leftspec-y&rightspec;
  do p = &leftspec to &rightspec;
    x(p) = (x(p) - specmean)/col1;
  end;
  rename col1=specstd;
  keep y&leftspec-y&rightspec chem1 numpoints numrecords specmean col1;
  run;

/* .......................................... 29 end ............................................ */

  %do i=&detrendbegin %to &detrendend;

/* .......................................... 30 start .......................................... */

    %if &i = 0 %then %do;

      data afterSNV;
      set SNV_apply;
      mergevariable = 'a';
      keep y&leftspec-y&rightspec chem1 numpoints numrecords mergevariable;
      run;

      data vafterSNV;
      set vSNV_apply;
      mergevariable = 'a';
      keep y&leftspec-y&rightspec chem1 numpoints numrecords mergevariable;
      run;

    %end;

    %else %if &i = 1 %then %do;

      proc transpose data=SNV_apply out=t_SNV_apply;
      var y&leftspec-y&rightspec;
      quit;

      data t_SNV_applyb;
      set t_SNV_apply;
      n = _n_;
      n_sq = n * n;
      keep col1-col&nrec n n_sq;
      run;

      proc reg data = t_SNV_applyb noprint;
      model col1-col&nrec = n n_sq;
      output out = detrend_apply  residual = resid1-resid&nrec;
      quit;

      proc transpose data=detrend_apply out=t_detrend_apply;
      var resid1-resid&nrec;
      quit;

      data detrendresid;
      set t_detrend_apply;
      array y(&leftspec:&rightspec) col&leftspec-col&rightspec;
      array z(&leftspec:&rightspec) z&leftspec-z&rightspec;
      do p = &leftspec to &rightspec;
        z(p) = y(p);
      end;
      keep z&leftspec-z&rightspec;
      run;

      data afterSNV;
      merge SNV_apply detrendresid;
      array y(&leftspec:&rightspec) y&leftspec-y&rightspec;
      array z(&leftspec:&rightspec) z&leftspec-z&rightspec;
      do l = &leftspec to &rightspec;
        y(l) = z(l);
      end;
      mergevariable = 'a';
      keep y&leftspec-y&rightspec chem1 numpoints numrecords mergevariable;
      run;

      proc transpose data=vSNV_apply out=vt_SNV_apply;
      var y&leftspec-y&rightspec;
      quit;

      data vt_SNV_applyb;
      set vt_SNV_apply;
      n = _n_;
      n_sq = n * n;
      keep col1-col&vnrec n n_sq;
      run;

      proc reg data = vt_SNV_applyb noprint;
      model col1-col&vnrec = n n_sq;
      output out = vdetrend_apply  residual = resid1-resid&vnrec;
      quit;

      proc transpose data=vdetrend_apply out=vt_detrend_apply;
      var resid1-resid&vnrec;
      quit;

      data vdetrendresid;
      set vt_detrend_apply;
      array y(&leftspec:&rightspec) col&leftspec-col&rightspec;
      array z(&leftspec:&rightspec) z&leftspec-z&rightspec;
      do p = &leftspec to &rightspec;
        z(p) = y(p);
      end;
      keep z&leftspec-z&rightspec;
      run;

      data vafterSNV;
      merge vSNV_apply vdetrendresid;
      array y(&leftspec:&rightspec) y&leftspec-y&rightspec;
      array z(&leftspec:&rightspec) z&leftspec-z&rightspec;
      do l = &leftspec to &rightspec;
        y(l) = z(l);
      end;
      mergevariable = 'a';
      keep y&leftspec-y&rightspec chem1 numpoints numrecords mergevariable;
      run;

    %end;

/* .......................................... 30 end ............................................ */

    %do j=&tablebegin %to &tableend;

/* .......................................... 31 start .......................................... */

      data chosentable;
      set savgol;
      if &j = 1 then convolutiontable = &firstSGtable;
      else if &j = 2 then convolutiontable = &secondSGtable;
      else convolutiontable = &thirdSGtable;
      if table ^= convolutiontable then delete;
      run;

      proc transpose data=chosentable out=t_chosentable;
      var c1-c11;
      quit;

/* .......................................... 31 end ............................................ */

      %do k=&windowbegin %to &windowend %by 2;

        %if
        (
        ^(&j = 1 and (&firstSGtable = 2 or &firstSGtable = 5 or &firstSGtable = 7) and &k = 5)
        and
        ^(&j = 2 and (&secondSGtable = 2 or &secondSGtable = 5 or &secondSGtable = 7) and &k = 5)
        and
        ^(&j = 3 and (&thirdSGtable = 2 or &thirdSGtable = 5 or &thirdSGtable = 7) and &k = 5)
        )
        %then %do; /* normal route, except when tables 2, 5, or 7 are chosen with 5 convolution pts */

/* .......................................... 32 start .......................................... */

          data convolute;
          set t_chosentable;
          if &j = 1 then table = &firstSGtable;
          else if &j = 2 then table = &secondSGtable;
          else table = &thirdSGtable;
          SG_pts = &k;
          rename col1=r1 col2=r2 col3=r3 col4=r4 col5=r5 col6=r6 col7=r7 col8=r8 col9=r9
                 col10=r10 col11=r11 col12=r12 col13=r13 col14=r14;
          if table = 3 or table = 4 or table = 5 or table = 9 then sign = -1; else sign = 1;
          mergevariable = 'a';
          keep sign SG_pts table col1-col14 mergevariable;
          if _n_ ^= 13.5 - SG_pts/2 then delete;
          run;

          data deriv;
          merge afterSNV convolute;
          by mergevariable;
          array y(&leftspec:&rightspec) y&leftspec-y&rightspec;
          array r(14) r1-r14;
          array z(&leftspec:&rightspec) z&leftspec-z&rightspec;
          newleft = (SG_pts+1)/2 + (&leftspec-1);
          newright = &rightspec - (SG_pts-1)/2;
          attrib charleft format = $CHAR.;
          charleft = newleft;
          charleft = compress(charleft);
          call symput('lefttrim',charleft);
          attrib charright format = $CHAR.;
          charright = newright;
          charright = compress(charright);
          call symput('righttrim',charright);
          do a = newleft to newright;
            sum = 0.0;
            do aa = (14 - (SG_pts+1)/2) to 13;
              sum = sum +        r(aa) * y(a-(13-aa));
            end;
            do aa = (14 - (SG_pts+1)/2) to 12;
              sum = sum + sign * r(aa) * y(a+(13-aa));
            end;
            z(a) = sum / r(14);
          end;
          keep z&leftspec-z&rightspec chem1 SG_pts table newleft newright numpoints numrecords mergevariable;
          run;

          data vderiv;
          merge vafterSNV convolute;
          by mergevariable;
          array y(&leftspec:&rightspec) y&leftspec-y&rightspec;
          array r(14) r1-r14;
          array z(&leftspec:&rightspec) z&leftspec-z&rightspec;
          newleft = (SG_pts+1)/2 + (&leftspec-1);
          newright = &rightspec - (SG_pts-1)/2;
          do a = newleft to newright;
            sum = 0.0;
            do aa = (14 - (SG_pts+1)/2) to 13;
              sum = sum +        r(aa) * y(a-(13-aa));
            end;
            do aa = (14 - (SG_pts+1)/2) to 12;
              sum = sum + sign * r(aa) * y(a+(13-aa));
            end;
            z(a) = sum / r(14);
          end;
          keep z&leftspec-z&rightspec chem1 SG_pts table newleft newright numpoints numrecords mergevariable;
          run;

/* .......................................... 32 end ............................................ */

          %do m = &varscalebegin %to &varscaleend;

/* .......................................... 33 start .......................................... */

            title6 "SNV with SG section. detrend_loop = &i, table_loop = &j, convwindow_loop = &k, varscale_loop = &m..";
            title7 "Program: &progname..  Output file: &SASstyleoutfile..";

            ods listing close;
            %if &m = 0 %then %do;

              proc pls data = deriv noscale cv = one nfac = &maxfac;
              model chem1 = z&lefttrim-z&righttrim;
              output out = plsout predicted = p_chem1;

            %end;

            %else %if &m = 1 %then %do;

              proc pls data = deriv cv = one nfac = &maxfac;
              model chem1 = z&lefttrim-z&righttrim;
              output out = plsout predicted = p_chem1;

            %end;

            ods output ResidualSummary=PRESS_results;
            ods output CVResults=crossval_results;

/* .......................................... 33 end ............................................ */

/* .......................................... 34 start .......................................... */

            data minerror1;
            set crossval_results;
            if _n_ = 1;
            rename nValue1=rootmeanPRESS;
            run;

            data minerror2;
            set crossval_results;
            if _n_ = 2;
            rename nValue1=optimal_factors;
            run;

            data minerror;
            merge minerror1 minerror2;
            run;

            title8 "The R-squared of the optimal model.";

            proc reg data = plsout corr;
            model p_chem1 = chem1;
            ods output Corr=REG_results;

            data getR;
            set REG_results;
            if _n_ = 2;
            R_sq = chem1 * chem1;
            run;

            data onelineout;
            %if &m = 0 %then %do;
              merge minerror getR;
            %end;
            %else %if &m = 1 %then %do;
              merge minerror getR chem1var;
              rootmeanPRESS = rootmeanPRESS * SQRT(v_chem1);
            %end;
            file "&SASstyleoutfile" mod;
            %if       &j = 1 %then %do;
              put @1 "SG_Table_= &firstSGtable" @;
            %end;
            %else %if &j = 2 %then %do;
              put @1 "SG_Table_= &secondSGtable" @;
            %end;
            %else %if &j = 3 %then %do;
              put @1 "SG_Table_= &thirdSGtable" @;
            %end;
            put @17 "SG_Window_= &k" @34 "SNV(0)_or_SNVdt(1)_= &i" @59
              "Noscale(0)_or_Variance_Scale(1)_= &m" @97 "MinPRESS_factors_=" @116 optimal_factors 2.0 @121
              "RootmeanPRESS_=" @137 rootmeanPRESS best9. @149 "R_sq_=" @156 R_sq 6.4;
            run;

            title8 "Summary of SAS Cross Validation Model, SNV section.";

            proc print data = onelineout;
            var Optimal_factors rootmeanPRESS R_sq;
            quit;

            proc means noprint data=PRESS_results;
            var RootMeanPRESS;
            output out = getthemin min = minRootMeanPRESS;
            quit;

            data minpressvalue;
            set getthemin;
            minPRESSValue = minRootMeanPRESS * minRootMeanPRESS * (&nrec - 1);
            mergevar = 'a';
            keep minPRESSValue mergevar;
            run;

            data addcommonvar;
            set PRESS_results;
            PRESSValue = RootMeanPRESS * RootMeanPRESS * (&nrec - 1);
            mergevar = 'a';
            if _n_ = 1 then delete;
            keep NumberOfFactors PRESSValue mergevar RootMeanPRESS;
            run;

            data gramsliketable;
            merge addcommonvar minpressvalue;
            by mergevar;
            F_Ratio = PRESSValue / minPRESSValue;
            F_Test = CDF('F',F_Ratio,(&nrec-1),(&nrec-1));
            minrec = 0;
            if F_Ratio = 1.0 then minrec = _n_;
            mergevarb = 'a';
            run;

            proc means noprint data = gramsliketable;
            var minrec;
            output out = theminrec max = minfac;
            quit;

            data addamergevar;
            set theminrec;
            mergevarb = 'a';
            run;

/* .......................................... 34 end .......................................... */

/* .......................................... 35 start .......................................... */

            %if &m = 0 %then %do;

              data gramstable;
              merge gramsliketable addamergevar;
              by mergevarb;
              if _n_ > minfac then do;
                F_Ratio = .;
                F_Test = .;
              end;
              run;
            %end;

            %else %if &m = 1 %then %do;

              data pregramstable;
              merge gramsliketable addamergevar;
              by mergevarb;
              if _n_ > minfac then do;
                F_Ratio = .;
                F_Test = .;
              end;
              run;

              data gramstable;
              merge pregramstable chem1var2;
              by mergevarb;
              PRESSValue = PRESSValue * v_chem1;
              run;

            %end;

/* .......................................... 35 end ............................................ */

/* .......................................... 36 start .......................................... */

            title7 "Program: &progname..  Output file: &Gramsstyleoutfile..";

            proc print noobs data = gramstable;
            var NumberOfFactors PRESSValue F_Ratio F_Test;
            quit;

            data trimgrams;
            set gramstable;
            if F_Test > 0.75 then delete;
            run;

            data gramsoptimal;
            set trimgrams;
            if _n_ > 1 then delete;
            run;

            data _null_;
            set gramsoptimal;
            attrib charoptfac format = $CHAR.;
            charoptfac = NumberOfFactors;
            charoptfac = compress(charoptfac);
            call symput('optfac',charoptfac);
            run;

            %if &m = 0 %then %do;

              proc pls data = deriv noscale nfac=&optfac;
              model chem1 = z&lefttrim-z&righttrim / solution;
              output out = plsout predicted = p_chem1;

            %end;

            %else %if &m = 1 %then %do;

              proc pls data = deriv nfac=&optfac;
              model chem1 = z&lefttrim-z&righttrim / solution;
              output out = plsout predicted = p_chem1;

            %end;
            
            ods output ParameterEstimates=bvector;

/* .......................................... 36 end ............................................ */

/* .......................................... 37 start .......................................... */

            proc reg data = plsout corr;
            model p_chem1 = chem1;
            ods output Corr=REG_results;

            data getR;
            set REG_results;
            if _n_ = 2;
            R_sq = chem1 * chem1;
            run;

            data bvectora;
            set bvector;
            If _n_ = 1 then do;
              attrib charbzero format = $CHAR.;
              charbzero = chem1;
              charbzero = compress(charbzero);
              call symput('bzero',charbzero);
              delete;
            end;
            run;

            proc transpose data=bvectora out=t_bvectora;
            var chem1;
            id RowName;
            quit;

            data t_bvectorb;
            set t_bvectora;
            _TYPE_ = 'SCORE';
            run;

            proc score data=vderiv score=t_bvectorb out=valout;
            var z&lefttrim-z&righttrim;
            quit;

            data valouta;
            set valout;
            predval = chem12 + &bzero;
            differ = predval - chem1;
            keep predval chem1 differ;
            run;

            proc means noprint data=valouta;
            var differ;
            output out = vstats mean = vBias std=vSEP uss=vUSS n=numsam;
            quit;

            proc reg data = valouta corr;
            model predval = chem1;
            ods output Corr=vREG_results;

            data vgetR;
            set vREG_results;
            if _n_ = 2;
            vr_sq = chem1 * chem1;
            run;

            data valstats;
            merge vstats vgetR;
            vRMSD = sqrt(vUSS/numsam);
            keep vBias vSEP vr_sq vRMSD;
            run;

            data onelineout;
            %if &m = 0 %then %do;
              merge gramsoptimal getR valstats;
              RootMeanPRESS = RootMeanPRESS * SQRT((&nrec-1)/&nrec);
            %end;
            %else %if &m = 1 %then %do;
              merge gramsoptimal getR chem1var valstats;
              RootMeanPRESS = RootMeanPRESS * SQRT((&nrec-1)/&nrec) * SQRT(v_chem1);
            %end;
            file "&Gramsstyleoutfile" mod lrecl = 325;
            rename RootMeanPRESS = RMSD;
            %if      &j = 1 %then %do;
              put @1 "SG_Table_= &firstSGtable" @;
            %end;
            %else %if &j = 2 %then %do;
              put @1 "SG_Table_= &secondSGtable" @;
            %end;
            %else %if &j = 3 %then %do;
              put @1 "SG_Table_= &thirdSGtable" @;
            %end;
            put @17 "SG_Window_= &k" @34 "SNV(0)_or_SNVdt(1)_= &i" @59
              "Noscale(0)_or_Variance_Scale(1)_= &m" @97 "Facs_at_min_error_=" @117 minfac 2.0 @122
              "Grams_factors_=" @138 NumberOfFactors 2.0 @143 "PRESS_Value_=" @157 PRESSValue best10.
              @170 "F_Ratio_=" @180 F_Ratio best8. @191 "F_Test_=" @200 F_Test best7. @210
              "RMSD_=" @217 RootMeanPRESS best11. @231 "R_sq_=" @238 R_sq 6.4 @247 "vBias_=" @255 vBias best8.
              @266 "vSEP_=" @273 vSEP best11. @287 "vr_sq_=" @295 vr_sq 6.4 @304 "vRMSD_=" @312 vRMSD best11.;
            run;

            title8 "Summary of Grams-Like Cross Validation and Validation.";

            ods listing;
            proc print noobs data = onelineout;
            var NumberOfFactors minfac PRESSValue RMSD F_Ratio F_Test R_sq vBias vSEP vr_sq vRMSD;
            quit;

/* .......................................... 37 end .......................................... */

          %end; /* m, variance scaling loop */

        %end; /* the normal route, except when table is 2, 5, or 7 and conv. pts = 5 */

      %end; /* k, SG convolution window loop */

    %end; /* j, SG table loop */

  %end; /* i, SNV detrend loop */

/* .......................................... 38 start .......................................... */

  %if &detrendbegin = 1 | &detrendend = 1 %then %do;

    proc datasets;
    delete
      t_before_msc t_before_msca mspectrum t_mspectrum_mean addmean t_mspectrum_std SNV_apply
      t_SNV_apply t_SNV_applyb detrend_apply t_detrend_apply detrendresid
      afterSNV chosentable t_chosentable convolute deriv
      plsout PRESS_results crossval_results minerror1 minerror2 minerror
      REG_results getR onelineout getthemin minpressvalue addcommonvar gramsliketable
      theminrec addamergevar gramstable trimgrams gramsoptimal
      vt_before_msc vt_before_msca vmspectrum vt_mspectrum_mean vaddmean vt_mspectrum_std vSNV_apply
      vt_SNV_apply vt_SNV_applyb vdetrend_apply vt_detrend_apply vdetrendresid
      vafterSNV vderiv
      bvector bvectora t_bvectora t_bvectorb valout valouta
      vstats vREG_results vgetR valstats;
    run; quit;

  %end;
  %else %do;

    proc datasets;
    delete
      t_before_msc t_before_msca mspectrum t_mspectrum_mean addmean t_mspectrum_std SNV_apply
      afterSNV chosentable t_chosentable convolute deriv
      plsout PRESS_results crossval_results minerror1 minerror2 minerror
      REG_results getR onelineout getthemin minpressvalue addcommonvar gramsliketable
      theminrec addamergevar gramstable trimgrams gramsoptimal
      vt_before_msc vt_before_msca vmspectrum vt_mspectrum_mean vaddmean vt_mspectrum_std vSNV_apply
      vafterSNV vderiv
      bvector bvectora t_bvectora t_bvectorb valout valouta
      vstats vREG_results vgetR valstats;
    run; quit;

  %end;

  %if &varscalebegin = 1 | &varscaleend = 1 %then %do;

    proc datasets;
    delete
      pregramstable;
    run; quit;

  %end;

/* .......................................... 38 end ............................................ */

%MEND SNVloop;


/* Macro that deletes the remaining datasets */

%MACRO cleanslate;

/* .......................................... 38a start .......................................... */

proc datasets;
delete
  chem1var chem1var2 compress_spectra_with_chemical vcompress_spectra_with_chemical savgol;
run; quit;

/* .......................................... 38a end ............................................ */

%MEND cleanslate;


/* ------------------------------------------  BEGIN MAIN PROGRAM ---------------------------------------------------------- */

options pageno = 1;

/* .......................................... 39 start .......................................... */

/* Use the following procedure to suppress the SAS log, but
   only after you are confident program is running correctly. */
/*
filename junk dummy;
proc printto log = junk;
run;
*/

%getspec; /* macro call to get spectral data */

%getchem; /* macro call to get chemical data */

%getSG; /* macro call to get Savitzky-Golay convolution coefficients */

/* The following 4 macro calls invoke the various pretreatments and PLS regression. */

%MSCnoSGloop;

%SGloop;

%SNVnoSGloop;

%SNVloop;

%cleanslate;

proc printto; /* for redirecting SAS log information to default device or location */
run;

/* .......................................... 39 end ............................................ */

/* --------------------------------------------  END MAIN PROGRAM ---------------------------------------------------------- */

